#!/usr/bin/env python
"""
Downloads chromedriver.
"""
import os
import platform
import re
import stat
import subprocess
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import IO
from xml.etree import ElementTree
from zipfile import ZipFile

import requests
import structlog

from hukudo.exc import HukudoException

logger = structlog.get_logger()

ARCHITECTURES = ['linux64', 'mac64', 'mac64_m1', 'win32']
# LHS: https://docs.python.org/3/library/platform.html#platform.system
# No Mac M2 :>
PLATFORM2ARCH = {
    'Linux': 'linux64',
    'Darwin': 'mac64',
    'Windows': 'win32',
}
DEFAULT_EXE = 'google-chrome'

try:
    cache_dir = os.environ['REQUESTS_CACHE_DIR']
    import requests_cache

    requests_cache.install_cache(cache_dir)
except KeyError:
    pass
except ModuleNotFoundError:
    logger.error('REQUESTS_CACHE_DIR set, but requests-cache not installed. Try: `pip install requests-cache`')
    raise


class ParseError(HukudoException):
    pass


class ChromedriverDownloadError(HukudoException):
    pass


@dataclass
class Version:
    version: str
    tuple: tuple[int]
    major: int
    url: str = None

    @classmethod
    def parse(cls, version):
        t = tuple(int(x) for x in version.split('.'))
        return cls(
            version=version,
            tuple=t,
            major=t[0],
        )

    @classmethod
    def from_re_match(cls, regex, text):
        m = re.match(regex, text)
        if m is None:
            raise ParseError(f'Could not parse version from "{text}"')
        return cls.parse(m.group(1))

    def __lt__(self, other):
        return self.tuple < other.tuple

    def __str__(self):
        return self.version


def my_chrome_version(exe=DEFAULT_EXE) -> Version:
    out = subprocess.check_output([exe, '--version'], encoding='utf-8')  # e.g. 'Google Chrome 100.0.4896.127 \n'
    return Version.from_re_match(r'.*?([\d.]+).*?', out)


def my_arch():
    return PLATFORM2ARCH[platform.system()]


def driver_version(chromedriver: Path):
    out = subprocess.check_output([chromedriver, '--version'], encoding='utf-8')
    return Version.from_re_match(r'ChromeDriver ([\d.]+).*?', out)


def is_matching_my_major(chromedriver: Path, browser_exe=DEFAULT_EXE):
    return my_chrome_version(browser_exe).major == driver_version(chromedriver).major


class ChromedriverRepo:
    def __init__(self, arch, url='https://chromedriver.storage.googleapis.com/'):
        if arch not in ARCHITECTURES:
            raise ValueError(f'Invalid architecture {arch}')
        self.arch = arch
        self.url = url
        # RAII
        response = requests.get(url)
        response.raise_for_status()
        # strip default namespace https://stackoverflow.com/a/35165997/241240
        xml = re.sub(r"""\sxmlns=["'].+?["']""", '', response.text, count=1)
        self.root = ElementTree.fromstring(xml)

    def versions(self):
        """
        Find versions for given architecture. See ARCHITECTURES for valid values.
        """
        versions = []
        for e in self.root.findall('./Contents/Key'):
            key = e.text  # e.g. "99.0.4844.17/chromedriver_linux64.zip"
            match = re.match(fr'^([\d.]+)/chromedriver_{self.arch}.zip', key)
            if match:
                v = Version.parse(match.group(1))
                v.url = self.url + key
                versions.append(v)
        versions = sorted(versions)

        return versions

    def latest_version(self, major) -> Version:
        all_versions = self.versions()
        hits = sorted(v for v in all_versions if v.major == major)
        return hits[-1]


def download_zip(version: Version) -> IO[bytes]:
    log = logger.bind(version=version)
    log.info('downloading', url=version.url)
    response = requests.get(version.url)
    response.raise_for_status()
    zipfile = ZipFile(BytesIO(response.content))
    return zipfile.open('chromedriver')


def download_latest(target_dir: Path, force=False) -> [Version, Path]:
    if not target_dir.is_dir():
        raise ChromedriverDownloadError(f'not a directory: {target_dir}')

    arch = my_arch()
    repo = ChromedriverRepo(arch)
    chromedriver_version = repo.latest_version(my_chrome_version().major)

    target = target_dir / f'chromedriver-{chromedriver_version}'
    if target.exists():
        if force:
            target.unlink()
        else:
            raise FileExistsError(target)

    fd = download_zip(chromedriver_version)
    target.write_bytes(fd.read())
    target.chmod(target.stat().st_mode | stat.S_IXUSR)
    return chromedriver_version, target


def download_and_symlink(target_dir: Path, force=False, skip_symlink=False):
    if target_dir is None:
        target_dir = Path.cwd()
    if not target_dir.is_dir():
        raise ChromedriverDownloadError(f'not a directory: {target_dir}')

    try:
        version, path = download_latest(target_dir, force=force)
    except FileExistsError as e:
        raise ChromedriverDownloadError(f'file exists: {e}')

    if skip_symlink is False:
        os.chdir(target_dir)
        link = Path('chromedriver')
        link.unlink(missing_ok=True)
        link.symlink_to(path.relative_to(target_dir))

    return version, path



import httpx
import structlog

from .errors import GrafanaError, DashboardWriteError, CertificateError
from .models import Dashboard

logger = structlog.get_logger()


class Grafana:
    VIEWER = 'Viewer'
    EDITOR = 'Editor'
    ADMIN = 'Admin'
    ROLES = [VIEWER, EDITOR, ADMIN]

    @staticmethod
    def bearer_header(api_key):
        return {'Authorization': f'Bearer {api_key}'}

    @classmethod
    def from_basic_auth(cls, url, username, password, verify=None, cert=None):
        client = httpx.Client(
            base_url=url,
            auth=(username, password),
            http2=True,
            verify=verify,
            cert=cert,
        )
        return cls(client)

    @classmethod
    def from_api_key(cls, url, api_key, verify=None, cert=None):
        client = httpx.Client(
            base_url=url,
            headers=cls.bearer_header(api_key),
            http2=True,
            verify=verify,
            cert=cert,
        )
        return cls(client)

    def __init__(self, client: httpx.Client):
        self.client = client
        self.log = logger.bind(instance=self)

    def __str__(self):
        return str(self.client.base_url)

    def __repr__(self):
        return f'Grafana("{self}")'

    def health(self):
        try:
            r = self.client.get('/api/health')
        except httpx.ConnectError as e:
            if 'unable to get local issuer certificate' in str(e):
                raise CertificateError('Try setting GRAFANA_CLIENT_ROOT_CA.')
            raise
        except httpx.ReadError as e:
            if 'BAD_CERTIFICATE' in str(e):
                raise CertificateError('Try setting GRAFANA_CLIENT_CERT and …_KEY')
            raise
        if r.status_code != 200:
            raise GrafanaError('no 200 on /api/health')
        elif r.json()['database'] != 'ok':
            raise GrafanaError('database nok')
        else:
            return True

    def provision_api_key(self, name, role):
        """
        https://grafana.com/docs/grafana/latest/http_api/create-api-tokens-for-org/
        """
        if role not in self.ROLES:
            raise ValueError('invalid role')
        response = self.client.post('/api/auth/keys', json={'name': name, 'role': role})
        response.raise_for_status()
        return response.json()['key']

    def dashboards(self):
        result = []
        # https://grafana.com/docs/grafana/latest/http_api/folder_dashboard_search/
        r = self.client.get('/api/search?query=&starred=false')
        r.raise_for_status()
        search_results = r.json()
        uids = [sr['uid'] for sr in search_results if sr['type'] == 'dash-db']
        for uid in uids:
            d = self.client.get(f'/api/dashboards/uid/{uid}').json()['dashboard']
            result.append(Dashboard(d))
        return result

    def post_dashboard(self, dashboard_data):
        # force overwrite by setting id = null
        # https://grafana.com/docs/grafana/latest/http_api/dashboard/#create--update-dashboard
        if 'id' in dashboard_data:
            dashboard_data['id'] = None
        uid = dashboard_data['uid']
        log = self.log.bind(uid=uid)
        body = {
            'dashboard': dashboard_data,
            'overwrite': True,
        }
        response = self.client.post('/api/dashboards/db', json=body)
        if response.status_code != 200:
            log.error(response.text, uid=uid)
            raise DashboardWriteError(uid)
        log.debug('create dashboard', uid=uid)



#!/usr/bin/env python
from pathlib import Path

from setuptools import setup

# https://packaging.python.org/en/latest/guides/making-a-pypi-friendly-readme/
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(long_description=long_description, long_description_content_type='text/markdown')



import re
import requests


def _iter_names():
    response = requests.get('https://pypi.org/simple/')
    links = re.findall(r'href="(.+?)"', response.text)
    for link in links:
        yield link.removeprefix('/simple/').removesuffix('/')


def names() -> [str]:
    """
    Return all package names from pypi.org using the simple index
    """
    return list(_iter_names())



import logging
import sys

LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG']


def configure_structlog_dev(log_level='WARNING'):
    import structlog

    if isinstance(log_level, str):
        log_level = getattr(logging, log_level.upper())
    structlog.configure(
        processors=[
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(fmt="iso", utc=True),
            # structlog.processors.CallsiteParameterAdder([
            #     structlog.processors.CallsiteParameter.MODULE,
            #     structlog.processors.CallsiteParameter.FUNC_NAME,
            #     structlog.processors.CallsiteParameter.LINENO,
            # ]),
            structlog.dev.ConsoleRenderer(),
        ],
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(file=sys.stderr),
        cache_logger_on_first_use=False,
    )
